
#include "Board.h"
#include <algorithm>
#include <stdio.h>

#include "bighashmap.cc"
template class BigHashMap<Point, HomeBase>;
template class BigHashMap<unsigned, Robot>;
template class BigHashMap<Point, Robot*>;
template class Matrix<char>;


Board::Board(Matrix<char>& m, unsigned id, Game& game) : 
  _board(m), _game(game)
{
  for (int i = 0; i < m._size.x; i++) {
    for (int j = 0; j < m._size.y; j++) {
      if (_board.at(i,j) == HOMEBASE) {
	Point p(i,j);
	_homebases.insert(p, HomeBase(p));
      }
    }
  }
}

void
Board::place_robot(unsigned id, Point pos) 
{
  _robots.insert(id, Robot(_game,id, pos));
  _robotPos.insert(pos, id);
}

void
Board::erase_robots(list<unsigned>&dead) {
  for(list<unsigned>::iterator i = dead.begin();
      i != dead.end(); i++) 
  {
    _robotPos.remove(_robots.findp(*i)->position());
    _robots.remove(*i);
  }
}

void
Board::erase_homebase(Point p) {
  if (_homebases.findp(p) != NULL)
    _homebases.remove(p);
}

/*
void
Board::visited(Point pos)
{
  HomeBase *hb = _homebases.findp(pos);
  if (hb)
    hb->_visited = true;
}
*/

/*
void
Board::set_packages(Point spot, const vector<unsigned> &packs)
{
  //  we only care about this to set homebase packages
  if (_board[spot.x][spot.y] != HOMEBASE)
    return;
  HomeBase* hb = _homebases.findp(spot);
  assert(hb);

  for (vector<unsigned>::const_iterator j = packs.begin(); 
      j != packs.end(); j++) 
    if (find(hb->_packages.begin(),hb->_packages.end(), *j) 
	!= hb->_packages.end())
      hb->_packages.push_back(*j);
  hb->_packages.sort();
}
*/

void
Board::update(GameMove& move) {
  vector<unsigned> anno;
  push(move, anno);
}
void 
Board::push(GameMove &move, vector<unsigned> &anno)
{
  Robot* rob = _robots.findp(move._robot_id);
  if (!rob) {
    printf ("Coudln't find robot %d!\n", move._robot_id);
    return;
    assert (rob);
  }
  switch (move._command) {
    case NOOP:
      break;
    case NORTH:
    case SOUTH:
    case EAST:
    case WEST:
      move_robot(rob, move._command, anno);
      break;
    case PICK:
      {
	move._items.sort();
	list<unsigned> items = move._items;
	rob->_packages.merge(items);
      }
      break;
    case DROP:
      move._items.sort();
      sorted_remove(rob->_packages, move._items);
      break;
  }
}

void
Board::pop(const GameMove&move, const vector<unsigned> &anno) 
{
  switch(move._command){
    case NOOP:
      break;
    case NORTH:
      for (vector<unsigned>::const_iterator i = anno.begin(); i != anno.end();
	  i++){
	_robots.findp(*i)->move(SOUTH);
      }
      break;
    case SOUTH:
      for (vector<unsigned>::const_iterator i = anno.begin(); i != anno.end();
	  i++){
	_robots.findp(*i)->move(NORTH);
      }
      break;
    case EAST:
      for (vector<unsigned>::const_iterator i = anno.begin(); i != anno.end();i++){
	_robots.findp(*i)->move(WEST);
      }
      break;
    case WEST:
      for (vector<unsigned>::const_iterator i = anno.begin(); i != anno.end();i++){
	_robots.findp(*i)->move(EAST);
      }
      break;
    case PICK:
      {
	Robot* rob = _robots.findp(move._robot_id);
	sorted_remove(rob->_packages, move._items);
      }
      break;
    case DROP:
      {
	Robot* rob = _robots.findp(move._robot_id);
	list<unsigned> items = move._items;
	rob->_packages.merge(items);
      }
      break;
  }
}

void
Board::sorted_remove(list<unsigned>& all, const list<unsigned>& removed)
{
  list<unsigned>::const_iterator i;
  list<unsigned>::iterator j;
  j = all.begin();
  for (i = removed.begin(); i != removed.end(); i++) {
    for (;j != all.end() && *j < *i; j++);
    if (*j == *i)
      j = all.erase(j);
  }
}

bool
Board::move_robot(Robot* rob, COMMAND c, vector<unsigned>&anno) {
  Point newloc(rob->position());
  switch (c) {
    case NORTH:
       newloc.y++;
      break;
    case SOUTH:
      newloc.y--;
      break;
    case EAST:
      newloc.x++;
      break;
    case WEST:
      newloc.x--;
      break;
    default:
      assert(0);
  }
  _robotPos.remove(rob->position());
  rob->set_position(newloc);
  _robotPos.insert(rob->position(), rob->_id);
  return true;

  /*
  if (try_move(newloc, c, anno)) {
    _robotPos.remove(rob->position());
    rob->set_position(newloc);
    _robotPos.insert(rob->position(), rob->_id);
    anno.push_back(rob->_id);
    return true;
  }
  return false;
  */
}

bool
Board::try_move(const Point&p, COMMAND c, vector<unsigned>&anno) {
  switch (_board.at(p.x,p.y)) {
    case WALL:
      return false;
    case WATER:
      return true;
    case HOMEBASE:
    case FLOOR:
      {
      unsigned *u =  _robotPos.findp(p);
      if (u == NULL)
	return true;
      return move_robot(_robots.findp(*u), c, anno);
      }
    default:
      assert(0);
  }
}

