#ifndef BOARD_H
#define BOARD_H

#include <vector>
#include "Types.h"
#include "bighashmap.hh"

/*
 * Board Spot Types:
 *
 * wall, water, homebase, recently-visited homebase, robot, 
 */
// moved to Types.h
/*#define WALL 0
#define WATER 1
#define HOMEBASE 2
#define FLOOR 4
*/
class GoalStrategy;
#include "Robot.h"

class Board
{
  public:
    Matrix<char>& _board;
    Game& _game;

    BigHashMap<Point, HomeBase> _homebases;
    BigHashMap<unsigned, Robot> _robots;
    BigHashMap<Point, unsigned> _robotPos;

    // returns false if blocked
    bool try_move(const Point&, COMMAND c, vector<unsigned> &anno);
    bool move_robot(Robot*, COMMAND c, vector<unsigned> &anno);

    void sorted_remove(list<unsigned>& all, const list<unsigned>& removed);

    friend GoalStrategy;

  public:
    Board(Matrix<char>&, unsigned id, Game&);
    Point size() const { return _board._size; }
    bool walkable(const Point&p) const { return _board.at(p.x,p.y) > WATER; }
    Point get_position(unsigned rid) const 
    	{ return _robots.findp(rid)->position(); }
    void erase_homebase(Point p);

    void place_robot(unsigned id, Point pos);
    void erase_robots(list<unsigned> &dead);
//    void set_packages(Point spot, const vector<unsigned> &packs);
    void update(GameMove&);
    void push(GameMove& move, vector<unsigned> &annotation);
    void pop(const GameMove& move, const vector<unsigned> &annotation);

};


#endif
