#include "Board.h"
#include "Distances.h"
#include "Types.h"
#include <queue>

template class Matrix<int>;
template class Matrix<char>;

Distances::Distances(Board& board)
  :
 _distances(board._board._size),
 _directions(board._board._size),
 _x(board._board._size.x),
 _y(board._board._size.y),
 _size(board._board._size),
 _origin(0,0),
 _board(board)
{
}

Distances::~Distances()
{
}

int
Distances::get_distance_to(const Point& p)
{
  return _distances.at(p);
}

int
Distances::get_direction_for(const Point& p)
{
  return _directions.at(p);
}


void
Distances::load_distances(const Point& origin)
{
  _origin = origin;
  queue<Point> Q;
  
  _distances.set_all_to(-1);
  _directions.set_all_to(DROP);
  
  _distances.at(origin) = 0;
  _directions.at(origin) = DROP;

  if (origin.y > 0) {
    Point s(origin.x,origin.y-1);
    if (_board.walkable(s)) {
      _distances.at(s) = 1;
      _directions.at(s) = SOUTH;
      Q.push(s);
    }
  }

  if (origin.y < _y-1) {
    Point n(origin.x,origin.y+1);
    if (_board.walkable(n)) {
      _distances.at(n) = 1;
      _directions.at(n) = NORTH;
      Q.push(n);
    }
  }

  if (origin.x > 0) {
    Point w(origin.x-1,origin.y);
    if (_board.walkable(w)) {
      _distances.at(w) = 1;
      _directions.at(w) = WEST;
      Q.push(w);
    }
  }

  if (origin.x < _x-1) {
    Point e(origin.x+1,origin.y);
    if (_board.walkable(e)) {
      _distances.at(e) = 1;
      _directions.at(e) = EAST;
      Q.push(e);
    }
  }


  while (!Q.empty()) {
    Point p = Q.front();
    Q.pop();
    int dist = _distances.at(p);
    int dir = _directions.at(p);

    if (p.y > 0) {
      Point so(p.x,p.y-1);
      if (_board.walkable(so) && _distances.at(so) == -1) {
	_distances.at(so) = dist+1;
	_directions.at(so) = dir;
	Q.push(so);
      }
    }

    if (p.y < _y-1) {
      Point no(p.x,p.y+1);
      if (_board.walkable(no) && _distances.at(no) == -1) {
	_distances.at(no) = dist+1;
	_directions.at(no) = dir;
	Q.push(no);
      }
    }

    if (p.x > 0) {
      Point we(p.x-1,p.y);
      if (_board.walkable(we) && _distances.at(we) == -1) {
	_distances.at(we) = dist+1;
	_directions.at(we) = dir;
	Q.push(we);
      }
    }

    if (p.y < _y-1) {
      Point ea(p.x+1,p.y);
      if (_board.walkable(ea) && _distances.at(ea) == -1) {
	_distances.at(ea) = dist+1;
	_directions.at(ea) = dir;
	Q.push(ea);
      }
    }
  }
  
}

