#ifndef GAME_H
#define GAME_H

/*
 * Board Spot Types:
 *
 * wall, water, homebase, recently-visited homebase, robot, 
 */

#include <vector>
#include "Stats.h"
#include "GoalStrategy.h"
#include "MoveStrategy.h"
#include "Distances.h"
#include "TravelTime.h"
#include "Types.h"
#include "Board.h"
#include "Player.h"
#include "Stats.h"
#include "Robot.h"

class Game
{
  private:
    // information about the current state
    Board _board;

    list<unsigned> _robots;
    list<unsigned> _deadRobots;
    list<unsigned> _standingOn; // will be replaced by mobile homebases
    BaggedPackage _packages[10001];

    Player _me;


    // objects/methods for planning
    GoalStrategy _gstrategy;
    MoveStrategy _mstrategy;
    Stats _stats;
    unsigned new_homebase_points();

    // temporary data structures
    Distances _distances;
    TravelTime _ttime;
    Point _cached_goal;
    list<unsigned> _packages_goal;

    friend GoalStrategy;
    friend MoveStrategy;
    friend Stats;
    friend Robot;
    friend TravelTime;

  public:
    Game(Matrix<char>& board, unsigned my_id, unsigned weight, 
	unsigned dollars, const Point&);
    void place_robot(unsigned id, Point pos);
    MyMove get_move(); // main function
    void update(GameMove&);
    void set_packages(list<unsigned>& packages);
    void set_package_specs(unsigned id, Point pos, unsigned weight);
    void print_board (ostream&);
};

#endif
