#ifndef GOALSTRATEGY_H
#define GOALSTRATEGY_H

class Game;
#include "Types.h"
#include <vector>
#include "bighashmap.hh"
#include "Player.h"


typedef pair<Point,Point> Coord;
int hashcode(const Coord&);

class DistanceVal {
  public:
  Coord _c;
  unsigned _dist;

  DistanceVal() {}
  DistanceVal(Coord c, unsigned d) : _c(c), _dist(d) {}

  bool operator==(const DistanceVal& dv) const { return _c == dv._c; }
};

class GoalStrategy
{
  private:
    Game& _game;
    int _depth;

    vector<unsigned> _packages;
    vector<Point> _points;
    BigHashMap<Coord, vector<DistanceVal> > _distancemap;

    void calculate_distances();
    void choose_packages();
    void choose_goal(const Player&, Point& goal, unsigned& score);
    	// choose-goal does NOT choose your current pos ever.
    unsigned heuristic(const Player &player);
    unsigned heuristic_base(const Player &player);

    unsigned get_distance(Point p1, Point p2);
    void get_random_subset(list<unsigned> &packages, unsigned maxweight);
    int choose_random_package(const vector<unsigned> &allpacks,
	const vector<bool> &in, unsigned weight, unsigned maxweight);
    int choose_nearby_package(const vector<unsigned> &allpacks,
	const vector<bool> &in, unsigned weight, unsigned maxweight,
	Point sum, int numItems);
    void complete_set(Point pos, const vector<unsigned> &allpacks,
	vector<bool> &in, unsigned &weight, unsigned maxweight,
	Point &sum, int numItems);

  public:
    GoalStrategy(Game&);
    void set_depth(int); // depth to search to, to adjust speed

    void get_goal(); // Where you'd like to go generally
};


#endif
