/*
 * @(#) $Id: Random.cpp,v 1.1 2002/09/01 23:33:57 ben Exp $
 * simple implementation of a random number generator from numerical receipts
 */

#include <math.h>
#include "Random.h"

/* constants for the function ran1 */
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

void seed_rand() {
        sran1(time(0) + getpid());
}
void seed_rand(int n) {
        sran1(n);
}
// returns a int in [0, max - 1] if max > 0.
u_int rand_int(u_int max) {
            return (u_int)(ran1() * max);
}
double rand_double() {
        return ran1();
}


// static var, stores the seed
static RandomGenerator g_random;

void sran1(int seed) { g_random.sran1(seed); }
double ran1() { return g_random.ran1(); }
double gasdev() { return g_random.gasdev(); }

void 
RandomGenerator::sran1(int seed) {
    if (seed < 0) {
	_ran_seed = (long) seed;
    } else if (seed == 0) {
	_ran_seed = -1;
    } else {
	_ran_seed = - seed;
    }
}

// static helper methods
static double _ran1(long *idum) {
    int j;
    long k;
    static long iy = 0;
    static long iv[NTAB];
    double temp;

    if (*idum <= 0 || !iy) {
	if (-(*idum) < 1) *idum = 1;    /* prevent *idum = 0 */
	else *idum = -(*idum);
	for (j = NTAB + 7; j >= 0; j--) {
	    k = (*idum)/IQ;
	    *idum = IA*(*idum - k * IQ) - IR * k;
	    if (*idum < 0) *idum += IM;
	    if (j < NTAB) iv[j] = *idum;
	}
	iy = iv[0];
    }
    k = (*idum) / IQ;
    *idum = IA * (*idum-k*IQ) - IR*k;
    if (*idum < 0) *idum += IM;
    j = iy/NDIV;
    iy = iv[j];
    iv[j] = *idum;
    if ((temp = AM*iy) > RNMX) return RNMX;
    else return temp;
}

double RandomGenerator::ran1(void) {
    return _ran1(&_ran_seed);
}

// static helper methods
static double _gasdev(long *idum) {
    double _ran1(long *idum);
    static int iset = 0;
    static double gset;
    double fac, rsq, v1, v2;

    if (iset == 0) {
	do {
	    v1 = 2.0 * _ran1(idum) - 1.0;
	    v2 = 2.0 * _ran1(idum) - 1.0;
	    rsq= v1*v1 + v2*v2;
	}
	while (rsq >= 1.0 || rsq == 0.0);

	fac = sqrt(-2.0 * log(rsq) / rsq);
	gset = v1*fac;
	iset=1;
	return v2*fac;
    } else {
	iset=0;
	return gset;
    }
}

double RandomGenerator::gasdev(void) {
    return _gasdev(&_ran_seed);
}

/* undefine the constants for the function ran1 */
#undef IA
#undef IM
#undef AM
#undef IQ
#undef IR
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
