#include "Robot.h"
#include "Game.h"
#include "Types.h"
#include <queue>

Robot::Robot()
  : _id(0),
  _distances(NULL),
  _position(Point(0,0)),
  _game(NULL),
  _dummy(true)
{}

Robot::Robot(Game& game, unsigned id, const Point& p, bool dummy)
  : _id(id),
  _distances(NULL), 
  _position(p),
  _game(&game),
  _dummy(dummy)
{
}

Robot::Robot(const Robot& r)
  : _id(r._id),
  _distances(NULL), 
  _position(r._position),
  _game(r._game), 
  _dummy(r._dummy)
{
}

Robot&
Robot::operator=(const Robot& r)
{
  if (&r != this) {
    _distances = NULL;
    _id = r._id;
    _position = r._position;
    _dummy = r._dummy;
  }
  return *this;
}

Robot::~Robot()
{
  if (_distances) delete(_distances);
}

void
Robot::move(COMMAND c)
{

}

int
Robot::get_distance_to(const Point& p) {

  if (_game && !_distances) {
    int memory = _game->_board._robots.size() * sizeof(int) *
      _game->_board._board._x * _game->_board._board._y;
    if (memory < 20000000) { // if we use less than 20 megs
      _distances = new Matrix<int>(_game->_board._board._size);
    }
  }
  
  if (_distances && _game) {
    if (_position != _cached) {
      calculate_distances();
    }

    return _distances->at(p);

  } else {
    int width = p.x - _position.x;
    int height = p.y - _position.y;
    if (width < 0) width *= -1;
    if (height < 0) height *= -1;
    return width + height;
  }
}

void
Robot::set_position(const Point& p)
{
  _position = p;
  /*
  if (_game && !_distances && !_dummy) {
    int memory = _game->_board._robots.size() * sizeof(int) *
      _game->_board._board._x * _game->_board._board._y;
    if (memory < 20000000) { // if we use less than 20 megs
      _distances = new Matrix<int>(_game->_board._board._size);
    }
  }

  if (_distances && !_dummy)
    calculate_distances();
    */
}

const Point&
Robot::position()
{
  return _position;
}

void
Robot::calculate_distances()
{
  if (!_game) return;
  queue<Point> Q;
  Point origin = _position;
  _cached = _position;
  Board *board = &_game->_board;
  int _x = _distances->_x;
  int _y = _distances->_y;
  

  _distances->set_all_to(-1);

  _distances->at(origin) = 0;

  if (origin.y > 0) {
    Point s(origin.x,origin.y-1);
    if (board->walkable(s)) {
      _distances->at(s) = 1;
      Q.push(s);
    }
  }

  if (origin.y < _y-1) {
    Point n(origin.x,origin.y+1);
    if (board->walkable(n)) {
      _distances->at(n) = 1;
      Q.push(n);
    }
  }

  if (origin.x > 0) {
    Point w(origin.x-1,origin.y);
    if (board->walkable(w)) {
      _distances->at(w) = 1;
      Q.push(w);
    }
  }

  if (origin.x < _x-1) {
    Point e(origin.x+1,origin.y);
    if (board->walkable(e)) {
      _distances->at(e) = 1;
      Q.push(e);
    }
  }

  while (!Q.empty()) {
    Point p = Q.front();
    Q.pop();
    int dist = _distances->at(p);

    if (p.y > 0) {
      Point so(p.x,p.y-1);
      if (board->walkable(so) && _distances->at(so) == -1) {
	_distances->at(so) = dist+1;
	Q.push(so);
      }
    }

    if (p.y < _y-1) {
      Point no(p.x,p.y+1);
      if (board->walkable(no) && _distances->at(no) == -1) {
	_distances->at(no) = dist+1;
	Q.push(no);
      }
    }

    if (p.x > 0) {
      Point we(p.x-1,p.y);
      if (board->walkable(we) && _distances->at(we) == -1) {
	_distances->at(we) = dist+1;
	Q.push(we);
      }
    }

    if (p.y < _y-1) {
      Point ea(p.x+1,p.y);
      if (board->walkable(ea) && _distances->at(ea) == -1) {
	_distances->at(ea) = dist+1;
	Q.push(ea);
      }
    }
  }
}
