/*
 * $Id: Serial.cpp,v 1.1.1.1 2002/09/01 21:14:47 ben Exp $
 * Serial.cpp -- thom, 18 june 2001
 *     Data serialization.
 *
 */

//#include <netinet/in.h>
#include "Serial.h"
#include <assert.h>

#define STRING_STACK_SIZE 32



/*---------------------------------
 * DataOutputStream implementation
 *---------------------------------
 */

int
DataOutputStream::writeBool (bool b) 
{
    char ch;
    if (b) 
	ch = 1;
    else 
	ch = 0;
   return writeChar(ch);
}


int
DataOutputStream::writeChar (char ch) 
{
	if (writeBuffer (&ch, 1) != 1) return -1;
	return 0;
}

int
DataOutputStream::writeShort (short16 nShort) 
{
	nShort = htons(nShort);
	if (writeBuffer (&nShort, 2) != 2) return -1;
	return 0;
}
int
DataOutputStream::writeShort (ushort16 nShort) 
{
	nShort = htons(nShort);
	if (writeBuffer (&nShort, 2) != 2) return -1;
	return 0;
}
	
int
DataOutputStream::writeInt(int32 nInt)
{
	nInt = htonl(nInt);
	if (writeBuffer ((void*)&nInt, 4) != 4) return -1;
	return 0;
}
int
DataOutputStream::writeInt(uint32 nInt)
{
	nInt = htonl(nInt);
	if (writeBuffer ((void*)&nInt, 4) != 4) return -1;
	return 0;
}
int
DataOutputStream::writeLong(ulong64 nLong) 
{
#if defined (ACE_LITTLE_ENDIAN)
    nLong = ((nLong & 0x00000000000000FFU) << 56) |
            ((nLong & 0x000000000000FF00U) << 40) |
            ((nLong & 0x0000000000FF0000U) << 24) |
            ((nLong & 0x00000000FF000000U) << 8) |
            ((nLong & (((ulong64)0x000000FFU) << 32)) >> 8) |
            ((nLong & (((ulong64)0x0000FF00U) << 32)) >> 24) |
            ((nLong & (((ulong64)0x00FF0000U) << 32)) >> 40) |
            ((nLong & (((ulong64)0xFF000000U) << 32)) >> 56);
#endif
   if (writeWholeBuffer((void*)&nLong, 8) != 8) return -1;
   return 0;
}

int
DataOutputStream::writeVarLenBuffer (int cchLen, char* rgchData, 
	int timeouts, int* stop)
{
    assert(cchLen >= 0);

    if (cchLen == 0) {
	return writeInt(cchLen);
    }

    char* wholeBuf = (char*)malloc(4+cchLen);
    *((int32*)wholeBuf) = htonl(cchLen);
    memcpy(wholeBuf+4, rgchData, cchLen);

    int bytesWritten = writeWholeBuffer 
	((void*)wholeBuf, cchLen+4, timeouts, stop);

    free(wholeBuf);

    if (bytesWritten != (cchLen+4))
	return -1;
    else 
	return 0;
}

int
DataOutputStream::writeObject(Serializable &ser)
{
    return ser.writeObject(*this);
}
int
DataOutputStream::writeString(const tstring &s) 
{
    // TEMP -- this may help us not crash for longer
    int size = s.size();
    if (size > MAX_STRING_SIZE)
	return -1;
    
    if (writeInt(size) < 0) 
	return -1;
    if (size == 0) return 0;
    return writeWholeBuffer((void*)s.data(), size*sizeof(tchar));
}


int
DataOutputStream::writeVectorString (vector<tstring> &v) {
    if (writeInt(v.size()) < 0) return -1;
    vector<tstring>::iterator i;
    for (i = v.begin(); i != v.end(); i++) {
	if (writeString(*i) < 0) return -1;
    }
    return 0;
}

int
DataOutputStream::writeWholeBuffer (void* buf, int len, int timeouts,
	int* stop) {
    int wrote = 0;
    int total = 0;
    int zero = 0;
    int timeoutsInARow = 0;
    if (stop == NULL) stop = &zero;

    // FIXME: change the 5 to a #define or param or something...
    while (total < len && timeouts > 0 && timeoutsInARow < 5) {
	wrote = writeBuffer((char*)buf + total, len - total);

	if (wrote == -1 && errno == ETIME) {
	    timeouts--;
	    timeoutsInARow++;
	} else if (wrote <= 0) {
	    return total;
	} else {
            total += wrote;
	    timeoutsInARow = 0;
	}
	if (*stop)
	    return total;
    }	
    return total;
}
/*--------------------------------
 * DataInputStream implementation
 *--------------------------------
 */

int
DataInputStream::readBool (bool &b) 
{
    char ch;
    if (readChar(ch) < 0) return -1;
    if (ch == 0) {
	b = false;
    } else {
	b = true;
    }
    return 0;
}

int
DataInputStream::readChar (char &buf) 
{
    if (readBuffer (&buf, 1) != 1) return -1;
    return 0;
}
int
DataInputStream::readChar (unsigned char &buf) 
{
    if (readBuffer (&buf, 1) != 1) return -1;
    return 0;
}

int
DataInputStream::readShort (short16 &buf) 
{
    short16 n;
    if (readBuffer (&n, 2) != 2) return -1;
    buf = ntohs(n);
    return 0;
}
int
DataInputStream::readShort (ushort16 &buf) 
{
    ushort16 n;
    if (readBuffer (&n, 2) != 2) return -1;
    buf = ntohs(n);
    return 0;
}	
int
DataInputStream::readInt(int32 &buf)
{
    int32 n;
    if (readBuffer (&n, 4) != 4) return -1;
    buf = ntohl(n);
    return 0;
}

int
DataInputStream::readInt(uint32 &buf)
{
    uint32 n;
    if (readBuffer (&n, 4) != 4) return -1;
    buf = ntohl(n);
    return 0;
}


int
DataInputStream::readLong(ulong64 &nLong) 
{
   if (readWholeBuffer((void*)&nLong, 8) != 8) return -1;
#if defined (ACE_LITTLE_ENDIAN)
    nLong = ((nLong & 0x000000FFU) << 56) |
            ((nLong & 0x0000FF00U) << 40) |
            ((nLong & 0x00FF0000U) << 24) |
            ((nLong & 0xFF000000U) << 8) |
            ((nLong & (((ulong64)0x000000FFU) << 32)) >> 8) |
            ((nLong & (((ulong64)0x0000FF00U) << 32)) >> 24) |
            ((nLong & (((ulong64)0x00FF0000U) << 32)) >> 40) |
            ((nLong & (((ulong64)0xFF000000U) << 32)) >> 56);
#endif
   return 0;
}

int
DataInputStream::readVarLenBuffer(uint32 &nLen, char * &buf, 
	uint32 maxsize, int timeouts, int* stop)
{

    // problem: it may take them a while to compress the data and even
    //    send the first byte
    // solution: duplicate read-int here, and read the int with time-outs too
    uint32 n;
    buf = NULL;
    if (readWholeBuffer (&n, 4, timeouts, stop) != 4) return -1;
    nLen = ntohl(n);

    if (nLen > maxsize)
		return -1;
    if (nLen == 0) {
	buf = NULL;
	return 0;
    }
    buf = (char*)malloc (nLen);
    if (readWholeBuffer ((void*)buf, nLen, timeouts, stop) != (int)nLen) {
	nLen = 0;
	free(buf);
	buf = NULL;
	return -1;
    } else {
	return 0;
    }
}

int 
DataInputStream::readLine(char* line, int buflen) {
    int read = 0;
    while (read < buflen) {
	if (readChar(line[read]) < 0)
	    return -1;
	read++;
	if (line[read-1] == '\n')
	    return read;
    }
    return -1;
}


int
DataInputStream::readObject(Serializable &ser)
{
    return ser.readObject(*this);
}

// BUGBUG -- this is not unicode compatible!
int
DataInputStream::readString(tstring &s, uint32 maxsize) 
{
    tchar stackBuf[STRING_STACK_SIZE];
    tchar *heapBuf;
    uint32 len;
    
    if (readInt(len) < 0) 
	return -1;
    if (len > maxsize) 
		return -1;
    if (len == 0) {
        s.resize(0);
	return 0;
    }
    if(len > STRING_STACK_SIZE) {
	heapBuf = new tchar [len];
	if (readWholeBuffer((void*)heapBuf, len*sizeof(tchar)) < 0) {
	    delete [] heapBuf;
	    return -1;
	}
	s.assign(heapBuf, len);
	delete [] heapBuf;
    } else {
	if (readWholeBuffer((void*)stackBuf, len*sizeof(tchar)) < 0) 
	    return -1;
	s.assign(stackBuf, len);
    }
    return 0;
}

int
DataInputStream::readVectorString(vector<tstring> &v, uint32 maxvec, 
	uint32 maxstr) {
    uint32 len;
    if (readInt(len) < 0) 
	return -1;
    if (len > maxvec)
		return -1;
    v.resize(len);
    vector<tstring>::iterator i;
    for (i = v.begin(); i != v.end(); i++) {
	if (readString(*i, maxstr) < 0) return -1;
    }
    return 0;
}

int
DataInputStream::readWholeBuffer (void* buf, int len, int timeouts, int* stop,
	int timeoutsInARowLimit) {
    int read = 0;
    int total = 0;
    int zero = 0;
    int timeoutsInARow = 0;
    if (stop == NULL) stop = &zero;
    //int error_number;
    while (total < len && timeouts > 0 && timeoutsInARow < timeoutsInARowLimit) {
	read = readBuffer((char*)buf + total, len - total);

	if (read < 0 && errno == ETIME) 
	{

	    timeouts--;
	    timeoutsInARow++;
	} else if (read <= 0)  {
	    /*
	    error_number = errno;
	    printf("%d", error_number);
	    */
	    return total;
	} else {
	    total += read;
	    timeoutsInARow = 0;
	}
	if (*stop)
	    return total;
    }
    return total;
}

/*-----------------------------------
 * FileStreams
 */

FileStream::FileStream(FILE* fp) : m_fp(fp) {}

void
FileStream::close() {
	fclose(m_fp);
}

int
FileStream::readBuffer(void* buf, int len) {
	return fread(buf, 1, len, m_fp);
}
int
FileStream::writeBuffer(void* buf, int len) {
	return fwrite(buf, 1, len, m_fp);
}

#ifdef _WIN32
//include -- bad hack to get this to work in windows
#include "FXM/DownloadPackage.h"

int 
DataInputStream::readVectorPtr(vector<DownloadPackage*> &v) {
	int len;
	v.clear();
	bool success = true;
	if (readInt(len) < 0) 
	    return -1;
	for (int i = 0; i < len; i++) {
	    DownloadPackage *s =  new DownloadPackage();
	    if (s->readObject(*this)< 0) {
		delete s;
		success = false;
		break;
	    }
	    v.push_back(s);
	}
	if (success == false) {
	    vector<DownloadPackage*>::iterator j;
	    for (j = v.begin(); j != v.end(); j++) {
		delete (*j);
	    }
	    v.clear();
	    return -1;
	}
	return 0;
}
int 
DataOutputStream::writeVectorPtr(vector<DownloadPackage*> &v) {
	if (writeInt(v.size()) < 0)  return -1;
	vector<DownloadPackage*>::iterator i;
	for (i = v.begin(); i != v.end(); i++) {
	    if ((*i)->writeObject(*this) < 0) return -1;
	}
	return 0;
}
#endif

/*------------------------------------
 * CharArrayStreams
 *
 * I think these should use ACE memory pools for efficiency
 */


#define CHAR_ARRAY_START_SIZE 128

CharArrayOutputStream::CharArrayOutputStream ():
	m_nBufSize(CHAR_ARRAY_START_SIZE),
	m_nDataLen(0),
	m_ownArray(true)
{
	m_buf = (char*)malloc(CHAR_ARRAY_START_SIZE);
}

CharArrayOutputStream::CharArrayOutputStream (int start_size):
	m_nBufSize(start_size),
	m_nDataLen(0),
	m_ownArray(true)
{
	m_buf = (char*)malloc(start_size);
	if (m_buf == NULL) {
		m_buf = (char*)malloc(CHAR_ARRAY_START_SIZE);
		m_nBufSize = CHAR_ARRAY_START_SIZE;
	}
}

int
CharArrayOutputStream :: writeBuffer (void* pBuffer, int nLen) 
{
	if (!m_ownArray) return -1;
	int nNewLen = nLen + m_nDataLen;

	if (m_nBufSize < nNewLen) {
		m_nBufSize *= 2;
		if (m_nBufSize < nNewLen) m_nBufSize = nNewLen;
		m_buf = (char*)realloc (m_buf, m_nBufSize);
		if (m_buf == NULL) return 0;
	}
	memcpy ((void*)(m_buf + m_nDataLen), pBuffer, nLen);
	m_nDataLen = nNewLen;
	return nLen;
}

void
CharArrayOutputStream::getCharArray (int &len, char* &buffer)
{
	if (!m_ownArray) {
	    len = 0;
	    buffer = NULL;
	} else {
	    len = m_nDataLen;
	    buffer = m_buf;
	}
}
void
CharArrayOutputStream::getOwnershipCharArray (uint32 &len, char* &buffer)
{
	len = m_nDataLen;
	buffer = m_buf;
	m_ownArray = false;
}
void
CharArrayOutputStream::getCopyCharArray (int &len, char* &buffer)
{
	if (!m_ownArray) {
	    len = 0;
	    buffer = NULL;
	} else {
	    len = m_nDataLen;
	    buffer = (char*)malloc (len);
	    memcpy(buffer, m_buf, len);
	}
}

void
CharArrayOutputStream::reset() 
{
    m_nDataLen = 0;
    if (!m_ownArray) {
	m_nBufSize = CHAR_ARRAY_START_SIZE;
	m_buf = (char*)malloc(CHAR_ARRAY_START_SIZE);
	m_ownArray = true;
    }
}

CharArrayOutputStream :: ~CharArrayOutputStream ()
{
    if (m_ownArray)
	free(m_buf);
}

CharArrayInputStream::CharArrayInputStream(int start_size) : 
	m_nBufSize (start_size),
	m_nDataPointer(0),
	m_nDataLen(0)
{
    if (start_size == 0) {
	m_buf = NULL;
    } else {
	m_buf = (char*)malloc(start_size);
	if (m_buf == NULL) {
	    m_buf = (char*)malloc(CHARARRAYINPUTSTREAMSTARTSIZE);
	    m_nBufSize = CHARARRAYINPUTSTREAMSTARTSIZE;
	    assert(m_buf);
	}
    }
}

CharArrayInputStream::CharArrayInputStream
	(int nLen, void* buffer) :
	m_nBufSize (nLen),
	m_nDataPointer(0),
	m_nDataLen(nLen)
{
    if (nLen == 0) {
	m_buf = NULL;
    } else {
	m_buf = (char *)malloc(nLen);
	assert(m_buf);
	memcpy ((void*)m_buf, buffer, nLen);
    }
}

void
CharArrayInputStream::setCharArray (int nLen, void* buffer)
{
    if (m_nBufSize != nLen) {
	// different buffer size, must resize
	m_buf = (char *)realloc(m_buf, nLen);
	m_nBufSize = nLen;
    }

    assert(m_buf);
    memcpy ((void*)m_buf, buffer, nLen);
    m_nDataPointer = 0;
    m_nDataLen = nLen;
}

int
CharArrayInputStream::readBuffer (void* buffer, int nLen)
{
    int nRead;
    int nLeft = m_nDataLen - m_nDataPointer;
	
    if (nLen > nLeft)
	nRead = nLeft;
    else
	nRead = nLen;

    if (nRead == 0) return 0;
    memcpy ((void*)buffer, (void*)(m_buf + m_nDataPointer), nRead);
    m_nDataPointer += nRead;
    return nRead;
}

int
CharArrayInputStream::getSize() {
    return m_nDataLen - m_nDataPointer;
}
	
CharArrayInputStream :: ~CharArrayInputStream ()
{
    if (m_buf != NULL)
	free(m_buf);
}
