
#include "Types.h"
#include "Stats.h"
#include "Game.h"

#include "bighashmap.cc"

template class Matrix<unsigned short>;
template class Matrix<unsigned>;

// pass in the Point size from the original map Matrix
Stats::Stats(Point& size, Game& game)
  : _game(game)
{
  robot_visits = new Matrix<unsigned short>(size.x,size.y);
  resistance = new Matrix<unsigned>(size.x,size.y);
  
  robot_visits->set_all_to(0);
  resistance->set_all_to(100);

  moves = 0;
  points = 0;
}

// should be done right before we figure out our move
void
Stats::log_robot_positions()
{
  unsigned k;
  for (k = 0; k < robot_visits->_x*robot_visits->_y; k++) {
    robot_visits->_data[k] *= 9;
    robot_visits->_data[k] /= 10;
  }
  
  BigHashMap<Point,unsigned>::Iterator i(&_game._board._robotPos);
  for (; i; i++) {
    // if its an enemy
    if (i.value() != _game._me._id) {
      // log his position
      unsigned old = robot_visits->at(i.key());
      robot_visits->at(i.key()) += 100;
    }
  }
  moves += 1;
}

void 
Stats::score(int points) {
  points += points;
}
unsigned
Stats::points_per_hundred_dollars(){
  int value = 100 * points / moves;
  if (value < 5) return 5;
  else return value;
}

unsigned
Stats::get_res_at(const Point& p)
{
  return resistance->at(p);
}

void
Stats::calculate_resistances()
{
  unsigned x,y,visits,percent,res;
  for (y = 1; y < robot_visits->_y-1; y++) {
    for (x = 1; x < robot_visits->_x-1; x++) {
      visits = robot_visits->at(x,y);
      // calculate percentage
      percent = visits/7;
      // calculate resustance
      res = get_resistance(x,y,percent);
      resistance->at(x,y) = res;
    }
  }   
}

unsigned
Stats::get_resistance(unsigned x, unsigned y, unsigned percent)
{
  double r = 100.0;
  set_mults(percent);
  r *= _mult;
  r *= mult_at(x+1,y);
  r *= mult_at(x-1,y);
  r *= mult_at(x,y+1);
  r *= mult_at(x,y-1);
  return (int)(r+0.5);
}

double
Stats::mult_at (unsigned x, unsigned y)
{
  switch (_game._board._board.at(x,y)) {
    case WALL:
      return _wallmult;
    case WATER:
      return _watermult;
    default:
      return 1.0;
  }
}

void
Stats::set_mults(unsigned percent)
{
  switch (percent) {
    case 0:
      _mult = 1.0;
      _wallmult = 1.0;
      _watermult = 1.0;
      break;
    case 1 ... 3:
      _mult = 1.0;
      _wallmult = 1.1;
      _watermult = 1.2;
      break;
    case 4 ... 6:
      _mult = 1.1;
      _wallmult = 1.2;
      _watermult = 1.4;
      break;
    case 7 ... 10:
      _mult = 1.3;
      _wallmult = 1.5;
      _watermult = 1.7;
      break;
    case 11 ... 20:
      _mult = 2.0;
      _wallmult = 3.0;
      _watermult = 3.5;
      break;
    case 21 ... 40:
      _mult = 4.0;
      _wallmult = 8.0;
      _watermult = 10.0;
      break;
    case 41 ... 100:
    default:
      _mult = 10.0;
      _wallmult = 10.0;
      _watermult = 10.0;
      break;
    
  }
}
