#include "TravelTime.h"
#include "Types.h"
#include "Board.h"
#include "Game.h"
#include <list>

template class Matrix<int>;
template class Matrix<char>;

TravelTime::TravelTime(const Board& board)
  :
 _distances(board.size()),
 _directions(board.size()),
 _x(board.size().x),
 _y(board.size().y),
 _size(board.size()),
 _board(board),
 _stats(&board._game._stats)
{
}

TravelTime::~TravelTime()
{
}

int
TravelTime::get_distance_to(const Point& p)
{
  return _distances.at(p);
}

int
TravelTime::get_direction_for(const Point& p)
{
  return _directions.at(p);
}


void
TravelTime::load_distances(const Point& origin)
{
  list<Point> Q;
  
  _distances.set_all_to(-1);
  _directions.set_all_to(DROP);
  
  _distances.at(origin) = 0;
  _directions.at(origin) = DROP;

  Point s(origin.x,origin.y-1);
  if (_board.walkable(s)) {
    _distances.at(s) = _stats->get_res_at(s);
    _directions.at(s) = SOUTH;
    insert(Q,s);
  }

  Point n(origin.x,origin.y+1);
  if (_board.walkable(n)) {
    _distances.at(n) = _stats->get_res_at(n);
    _directions.at(n) = NORTH;
    insert(Q,n);
  }

  Point w(origin.x-1,origin.y);
  if (_board.walkable(w)) {
    _distances.at(w) = _stats->get_res_at(w);
    _directions.at(w) = WEST;
    insert(Q,w);
  }

  Point e(origin.x+1,origin.y);
  if (_board.walkable(e)) {
    _distances.at(e) = _stats->get_res_at(e);
    _directions.at(e) = EAST;
    insert(Q,e);
  }


  while (!Q.empty()) {
    Point p = Q.front();
    Q.pop_front();
    int dist = _distances.at(p);
    int dir = _directions.at(p);

    if (p.y > 0) {
      Point so(p.x,p.y-1);
      if (_board.walkable(so) && _distances.at(so) == -1) {
	_distances.at(so) = dist+_stats->get_res_at(so);
	_directions.at(so) = dir;
    	insert(Q,so);
      }
    }

    if (p.y < _y-1) {
      Point no(p.x,p.y+1);
      if (_board.walkable(no) && _distances.at(no) == -1) {
	_distances.at(no) = dist+_stats->get_res_at(no);
	_directions.at(no) = dir;
    	insert(Q,no);
      }
    }

    if (p.x > 0) {
      Point we(p.x-1,p.y);
      if (_board.walkable(we) && _distances.at(we) == -1) {
	_distances.at(we) = dist+_stats->get_res_at(we);
	_directions.at(we) = dir;
    	insert(Q,we);
      }
    }

    if (p.y < _y-1) {
      Point ea(p.x+1,p.y);
      if (_board.walkable(ea) && _distances.at(ea) == -1) {
	_distances.at(ea) = dist+_stats->get_res_at(ea);
	_directions.at(ea) = dir;
    	insert(Q,ea);
      }
    }
  }
  
}

void
TravelTime::insert(list<Point>& Q, Point p)
{
  list<Point>::iterator i;
  i = Q.begin();
  while (i != Q.end() &&_distances.at(*i) < _distances.at(p)) {
    i++;
  }
  Q.insert(i,p);
}
