#ifndef TYPES_H
#define TYPES_H

#include <list>
#define MAXSIZE 1002

#define WALL 0
#define WATER 1
#define HOMEBASE 2
#define FLOOR 4

class Point{
  public:
    unsigned short x;
    unsigned short y;
    Point() {}
    Point(const Point& p) {x = p.x; y = p.y;}
    Point( unsigned short X, unsigned short Y) : x(X), y(Y) {}
    /*
    bool operator==(const Point& p) 
    { return reinterpret_cast<unsigned>(*this) ==
      reinterpret_cast<unsigned>(p);}
    bool operator != (const Point &p)
    { return reinterpret_cast<unsigned>(*this) !=
    reinterpret_cast<unsigned>(p);}
    */
    ~Point() {}
    bool operator==(const Point& p) const {
      return (x == p.x && y == p.y);
    }
    bool operator!=(const Point& p) {
      return (x != p.x || y != p.y);
    }
	  
};

int hashcode(const Point &p1);
int hashcode(const unsigned &u);
/*
{
//  return reinterpret_cast<unsigned>(p1);
  return (p1.x << 16) + p1.y;
}
*/

class BaggedPackage
{
  public:
    BaggedPackage(unsigned id,Point d,unsigned w)
      : _id(id),_dest(d),_weight(w) {}
    BaggedPackage() : _id(0), _dest(0,0), _weight(0) {}
    unsigned _id;
    Point _dest;
    unsigned _weight;
};

class HomeBase
{
  public:
    HomeBase() : _position(0,0),  _last_value(-1), _visited(false)  {}
    HomeBase(Point p) : _position(p),  _last_value(-1), _visited(false) {}

    Point _position;
    list<unsigned> _packages;
    int _last_value;
    bool _visited;
};


typedef enum {NOOP, NORTH, SOUTH, EAST, WEST, PICK, DROP} COMMAND;
class MyMove {
  public:
    COMMAND _command;
    list<unsigned> _items; 	// for pick, drop
    unsigned _dollars;		// how much to pay
};

class GameMove {
  public:
    COMMAND _command;
    unsigned _robot_id;
    list<unsigned> _items;
};

template <class T>
class Matrix {
  public:
  T *_data;
  
  unsigned _x,_y;
  Point _size;
  
  Matrix(unsigned x, unsigned y) 
    : _x(x), _y(y), _size(x,y)
  {
    _data = (T*)malloc(_x*_y*sizeof(T));
  }
  Matrix(const Point& size)
    : _x(size.x), _y(size.y), _size(size)
  {
    _data = (T*)malloc(_x*_y*sizeof(T));
  }
  ~Matrix() { free(_data); }
  
  T& at(unsigned x,unsigned y) {
    //if (x < 0 || y < 0 || x >= _x || y >= _y)
    //  return _dummy;
    return _data[y*_x + x];
  }
  T& at(Point p) {
    return at(p.x,p.y);
  }
  
  void set_all_to(T v) {
    int i;
    for (i = 0; i < (int)(_x*_y); i++) {
      _data[i] = v;
    }
  }

  char char_at(unsigned x, unsigned y) {
    int thing = (int)at(x,y);
    switch (thing) {
      case WATER:
	return '~';
      case WALL:
	return '#';
      case FLOOR:
	return '.';
      case HOMEBASE:
	return '@';
      default:
	return '?';
    }
  } 
};

#endif
