#include "Common.h"
#include "SockStream.h"
#include <cstdio>
#include <iostream>
#include "Types.h"
#include "Game.h"
#include "unistd.h"
#include "ace/OS.h"

template class Matrix<char>;

char readInt(unsigned &value, SockStream& ss);
int readBoardSpot(SockStream& ss);
char readWhiteSpace(SockStream& ss);
char readPlayerLocation(unsigned& id, unsigned& x, unsigned& y, SockStream& ss);
char readCoord(unsigned&value, SockStream& ss);
char readPlayerID(unsigned& value, SockStream& ss);
char readPackage(unsigned& id, unsigned& weight, unsigned& x, unsigned& y, SockStream& ss);
char readChar(SockStream& ss);
char readMoves(vector<GameMove>& moves, vector<Robot>& nrobs, SockStream& ss,
    Game& game);

void writeInt(unsigned value, SockStream& ss);
void writeNewLine(SockStream& ss);
void writeSpace(SockStream& ss);
void writeMove(MyMove& move, SockStream& ss);

tstring getCommand(COMMAND c);
tstring itoa(int n);

int main (int argc, char **argv) {

  if (argc < 2) {
    cout << "Error - params are hostname port" << endl;
    exit(0);
  }

  tstring hostname = argv[1];
  int port = atoi(argv[2]);

  cout << "Connecting to server..." << endl;
  cout << "hostname: " << hostname << endl << "port: " << port << endl;

  ACE_INET_Addr addr((u_short)port,hostname.c_str());
  ACE_SOCK_Stream stream;
  ACE_SOCK_Connector connector(stream,addr);

  SockStream ss(stream);

  char *player = "Player\n";
  ss.writeBuffer((void*)player,7);

  unsigned x,y;
  readInt(x,ss);
  readInt(y,ss);

  Matrix<char> board(x+2,y+2);
  board.set_all_to(0);
  unsigned i,j;

  // read in the board
  for (i = 1; i <= y; i++) {
    for (j = 1; j <= x; j++) {
      board.at(j,i) = readBoardSpot(ss);
    }
    readWhiteSpace(ss);
  }

  // print the board for the world to see
  for (i = y+2; i > 0; i--) {
    for (j = 0; j < x+2; j++) {
      cout << board.char_at(j,i-1);
    }
    cout << endl;
  }

  unsigned id,weight,money;
  readInt(id,ss);
  readInt(weight,ss);
  readInt(money,ss);

  cout << "ID: " << id << endl;
  cout << "Weight: " << weight << endl;
  cout << "Money: " << money << endl;

  // MAKE GAME HERE
  Point size(board._size);
  Game game(board,id,weight,money,size);
  
  unsigned rid,rx,ry;
  while (readPlayerLocation(rid,rx,ry,ss) != '\n') 
  {
    cout << "Player #" << rid << " is at (" << rx << "," << ry << ")" 
      << endl;
    game.place_robot(rid,Point(rx,ry)); // PLACE ROBOT
  }
  cout << "Player #" << rid << " is at (" << rx << "," << ry << ")" 
    << endl;
  game.place_robot(rid,Point(rx,ry)); // PLACE ROBOT

  list<unsigned> packages;
  unsigned pid,pweight,px,py;
  //int counter = 0;

  while (true) {
    // READ AVAILABLE PACKAGES
    packages.clear();
    while (readPackage(pid,pweight,px,py,ss) != '\n') {
      packages.push_back(pid);
      cout << "PID: " << pid << " WEIGHT: " << pweight
	<< " HEADED TO: (" << px << "," << py << ")" << endl;
      game.set_package_specs(pid,Point(px,py),pweight); // SET PACKAGE
    }
    if (pid != UINT32_MAX) {
      packages.push_back(pid);
      cout << "PID: " << pid << " WEIGHT: " << pweight
	<< " HEADED TO: (" << px << "," << py << ")" << endl;
      game.set_package_specs(pid,Point(px,py),pweight); // SET PACKAGE
    }

    // report package list
    game.set_packages(packages);

    game.print_board(cerr);
    MyMove move = game.get_move();
    /*
    switch (counter % 4) {
      case 0:
	move._command = EAST;
	break;
      case 1:
	move._command = NORTH;
	break;
      case 2:
	move._command = WEST;
	break;
      case 3:
	move._command = SOUTH;
	break;
    }
    move._dollars = 200;
    counter++;
    */
    // WRITE MOVE
    writeMove(move,ss);

    vector<GameMove> moves;
    vector<Robot> newrobots;
    // READ WHAT HAPPENED THAT TURN
    moves.clear();
    readMoves(moves,newrobots,ss,game);

    vector<Robot>::iterator ri;
    for (ri = newrobots.begin(); ri != newrobots.end(); ri++) {
      game.place_robot((*ri)._id,(*ri).position());
    }

    vector<GameMove>::iterator gi;
    for (gi = moves.begin(); gi != moves.end(); gi++) {
      game.update(*gi);
    }    
  }
}

  void
writeMove(MyMove& move, SockStream& ss)
{
  list<unsigned>::iterator i;

  tstring movestr = "";
  movestr += itoa(move._dollars);
  movestr += " ";

  movestr += getCommand(move._command);

      movestr += " ";
  if ((move._command == PICK || move._command == DROP) &&
      move._items.size() > 0) {
    for (i = move._items.begin(); i != move._items.end(); i++) {
      movestr += " ";
      movestr += itoa(*i);
    }
  }
  movestr += "\n";
  cout << movestr << endl;
  ss.writeBuffer((void*)(movestr.c_str()),strlen(movestr.c_str()));
}

  tstring
getCommand(COMMAND c)
{
  switch (c) {
    case NOOP:
      return tstring("Do Nothing");
    case NORTH:
      return tstring("Move N");
    case SOUTH:
      return tstring("Move S");
    case EAST:
      return tstring("Move E");
    case WEST:
      return tstring("Move W");
    case PICK:
      return tstring("Pick");
    case DROP:
      return tstring("Drop");
  }
  return "AHHHHHH!";
}

  void
writeSpace(SockStream& ss)
{
  char buffer = ' ';
  if (ss.writeBuffer((void*)&buffer,1) < 0) {
    cout << "Couldn't write - guess we died" << endl;
    exit(0);
  }
}

  void
writeNewLine(SockStream& ss)
{
  char buffer = '\n';
  if (ss.writeBuffer((void*)&buffer,1) < 0) {
    cout << "Couldn't write - guess we died" << endl;
    exit(0);
  }
}

  void
writeInt(unsigned value, SockStream& ss)
{

  unsigned lowest = value % 10;
  if (value >= 10) {
    writeInt(value / 10, ss);
  }
  char c = (char)lowest + '0';
  if (ss.writeBuffer((void*)&c,1) < 0) {
    cout << "Couldn't write - guess we died" << endl;
    exit(0);
  }
}

  char
readMoves(vector<GameMove>& moves, vector<Robot>& newrobots, SockStream& ss,
    Game& game)
{
  char buffer;
  bool numing = false;
  GameMove move;
  Point p(0,0);
  Robot newrobot(game,0,p,true);
  unsigned temp = 0;

  while (true) {
    // read first char	
    if (ss.readBuffer((void*)&buffer,1) < 0) {
      cout << "We failed a read!" << endl;
      exit(0);
    }

    switch (buffer) {
      // player declaration
      case '#':
	buffer = readInt(temp,ss);
	move._robot_id = (unsigned)temp;
	move._command = NOOP;
	moves.push_back(move);
	cout << "Robot #" << temp << ":" << endl;
	temp = 0;
	if (buffer == '\n') return '\n';
	break;
	// direction
      case 'N':
	move._command = NORTH;
	moves.push_back(move);
	cout << "Moved North" << endl;
	if (readChar(ss) == '\n') return '\n';
	break;
      case 'E':
	move._command = EAST;
	moves.push_back(move);
	cout << "Moved East" << endl;
	if (readChar(ss) == '\n') return '\n';
	break;
      case 'W':
	move._command = WEST;
	moves.push_back(move);
	cout << "Moved West" << endl;
	if (readChar(ss) == '\n') return '\n';
	break;
      case 'S':
	move._command = SOUTH;
	moves.push_back(move);
	cout << "Moved South" << endl;
	if (readChar(ss) == '\n') return '\n';
	break;
	// picked up
      case 'P':
	move._command = PICK;
	cout << "Pick" << endl;
	readChar(ss); // required space
	break;
	// dropped
      case 'D':
	move._command = DROP;
	cout << "Drop" << endl;
	readChar(ss); // required space
	break;
	// coordinates
      case 'X':
	readChar(ss);
	readInt(temp,ss);
	p.x = (unsigned)temp;
	readChar(ss);
	readChar(ss);
	buffer = (readInt(temp,ss) == '\n');
	p.y = (unsigned)temp;
	newrobot.set_position(p);
	newrobot._id = move._robot_id;
	newrobots.push_back(newrobot);
	if (buffer == '\n') return '\n';
	temp = 0;
	break;
	// newline
      case '\n':
	if (numing) {
	  move._items.clear();
	  move._items.push_back(temp);
	  moves.push_back(move);
	}
	return '\n';
	break;
      case '0' ... '9':
	numing = true;
	temp *= 10;
	temp += (int)(buffer - '0');
	break;
	// finished reading in a number (pick, drop)
      case ' ':
	numing = false;
	move._items.clear();
	move._items.push_back(temp);
	moves.push_back(move);
	temp = 0;
	break;
    }
  }
}

  char
readPackage(unsigned& id, unsigned& weight, unsigned& x, unsigned& y, 
    SockStream& ss)
{
  if (readInt(id,ss) == '\n') {
    id = UINT32_MAX;
    return '\n';
  }
  readInt(x,ss);
  readInt(y,ss);
  return readInt(weight,ss);
}

  char
readPlayerLocation(unsigned& id, unsigned& x, unsigned& y, SockStream& ss)
{
  readPlayerID(id,ss);
  readCoord(x,ss);
  return readCoord(y,ss);
}

  char
readCoord(unsigned&value, SockStream& ss)
{
  char buffer;

  if (ss.readBuffer((void*)&buffer,1) < 0) {
    cout << "We failed a read!" << endl;
    exit(0);
  }

  readWhiteSpace(ss);
  return readInt(value,ss);
}

  char
readPlayerID(unsigned& value, SockStream& ss)
{
  char buffer;

  if (ss.readBuffer((void*)&buffer,1) < 0) {
    cout << "We failed a read!" << endl;
    exit(0);
  }
  if (buffer == '\n') return buffer;

  if (buffer != '#') {
    cout << "Ack! no # in player ID" << endl;
    exit(0);
  }

  return readInt(value,ss);
}

  char
readInt(unsigned &value, SockStream& ss) 
{

  int temp = 0;
  char buffer = '1';
  bool first = true;

  while (buffer >= '0' && buffer <= '9') {
    if (ss.readBuffer((void*)&buffer,1) <= 0) {
      cout << "We failed a read!" << endl;
      exit(0);
    }

    if (first && buffer == '\n')
      return buffer;
    first = false;
    //		cout << buffer;

    if (buffer >= '0' && buffer <= '9') {
      temp *= 10;
      temp += (int)(buffer - '0');
    }
  }
  value = temp;
  return buffer;
}

  int
readBoardSpot(SockStream& ss)
{
  char buffer;

  if (ss.readBuffer((void*)&buffer,1) < 0) {
    cout << "We failed a read!" << endl;
    return -1;
  }

  //	cout << buffer;

  switch (buffer) {
    case '.':
      return FLOOR;
    case '#':
      return WALL;
    case '@':
      return HOMEBASE;
    case '~':
      return WATER;
    default:
      cout << "Error occured reading board spot" << endl;
      exit(0);
  }
}

  char
readWhiteSpace(SockStream& ss) 
{
  char buffer;

  if (ss.readBuffer((void*)&buffer,1) < 0) {
    cout << "We failed a read!" << endl;
    return -1;
  }

  //	cout << buffer;

  if (!isspace(buffer)) {
    cout << "wasn't whitespace!!!" << endl;
    exit(0);
  }

  return buffer;
}

  char
readChar(SockStream& ss) 
{
  char buffer;

  if (ss.readBuffer((void*)&buffer,1) < 0) {
    cout << "We failed a read!" << endl;
    return -1;
  }

  return buffer;
}

  void
reverse(tchar s[])
{
  int c, i, j;
  for (i = 0, j = t_strlen(s) - 1; i < j; i++, j--) {
    c = s[i];
    s[i] = s[j];
    s[j] = c;
  }
}

tstring
itoa(int n) {
  tchar s[13];
  int i, sign;
  if ((sign = n) < 0)
    n = -n;
  i = 0;
  do {
    s[i++] = n % 10 + '0';
  } while ((n /= 10) > 0);
  if (sign < 0)
    s[i++] = '-';
  s[i] = '\0';
  reverse(s);
  return tstring(s);
}
